#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>

#include <assert.h>
#include <math.h>

#include "debug.h"
#include "complex.h"

#define min(a, b) (((a) < (b)) ? (a) : (b))

#define SHARP (-1)

#define Unit (Val_int (0))

#define REPULSE 0.2
#define STABLE_LENGTH 1.0
#define ATTRACT 1.0
#define VISCOSITY 2.0
#define RUBBING 10.0

#define DT 0.1

#define BORDER 20

static value G; /* le graphe a dessiner */
static int n, d; /* la taille du graphe */

static cplx_t *pos, *speed, *accel;
static double x_min, x_max, y_min, y_max;
static double v_max;

static int width, height;

/* variables Caml a exporter */
static value int_pos;

/* le k-ieme voisin de node */
int neigbhoor (int node, int k)
{  
  return Int_val (Field (Field (G, node), k)); 
}



double node_repulse (double dist)
{  return exp (-REPULSE * dist); }

double edge_attract (double dist)
{  return (dist - STABLE_LENGTH) * ATTRACT; }

double rubbing (double speed)
{ return speed * VISCOSITY; }


void print (cplx_t *tab)
{
  int i;
  return;

  for (i = 0; i < n; i++) {
    printf ("%d : (%g, %g)\n", i, tab[i].re, tab[i].im);
  }
}

void print_int_pos (void)
{
  int i; return;

  for (i = 0; i < n; i++) {
    printf ("%d : (%d, %d)\n",
	    i,
	    Int_val (Field (Field (int_pos, i), 0)),
	    Int_val (Field (Field (int_pos, i), 1)));
  }
}

void init_state (void)
{
  int k;
  debug_in (0, "Initializing positions...");
  for (k = 0; k < n; k++) {
    ni_root (pos + k, n, k);
    //    cr_mul (pos + k, (double) n / 2);
    accel[k].re = accel[k].im = 0.0;
    speed[k].re = speed[k].im = 0.0;
  }
  debug_out (0, "Done.");
}

void compute_int_pos (void)
{
  int i;
  double ratio = min (((double) (width - 2 * BORDER)) / (x_max - x_min),
		      ((double) (height - 2 * BORDER)) / (y_max - y_min));
  double ax = (((double) width) - ratio * (x_max - x_min)) / 2.0;
  double ay = (((double) height) - ratio * (y_max - y_min)) / 2.0;

  debug_in (0, "Computing integer positions...");
  for (i = 0; i < n; i++) {
    Store_field (Field (int_pos, i),
		 0,
		 Val_int ((int) (ax + ratio * (pos[i].re - x_min))));
    Store_field (Field (int_pos, i),
		 1,
		 Val_int ((int) (ay + ratio * (pos[i].im - y_min))));
  }
  debug_out (0, "Done.");
}

/* graph -> int -> int-> (int * int) array */
value caml_init (value graph, value w, value h) 
{
  int i;
  CAMLparam1 (graph);

  set_debug_level (-1);
  debug_in (0, "Initializing...");

  register_global_root (&G);
  G = Field (graph, 0);
  n = Int_val (Field (graph, 1));
  d = Int_val (Field (graph, 2));

  width = Int_val (w);
  height = Int_val (h);

  pos = malloc (sizeof (cplx_t) * n);
  assert (pos != NULL);
  speed = malloc (sizeof (cplx_t) * n);
  assert (speed != NULL);
  accel = malloc (sizeof (cplx_t) * n);
  assert (accel != NULL);

  print (pos);
  init_state ();
  print (pos);
  x_min = y_min = -1.0;
  x_max = y_max = 1.0;

  register_global_root (&int_pos);
  int_pos = alloc_tuple (n);
  for (i = 0; i < n; i++) {
    Store_field (int_pos, i, alloc_tuple (2));
  }

  compute_int_pos ();
  print_int_pos ();

  debug_out (0, "Done.");
  CAMLreturn (int_pos);
}

void compute_pos (void)
{
  cplx_t tmp;
  int i;
  debug_in (0, "Computing positions...");
  for (i = 0; i < n; i++) {
    /* pos <- pos + speed * dt */
    tmp = speed[i];
    cr_mul (&tmp, DT);
    cc_add (pos + i, &tmp);
    if (pos[i].re < x_min) x_min = pos[i].re;
    if (pos[i].re > x_max) x_max = pos[i].re;
    if (pos[i].im < y_min) y_min = pos[i].im;
    if (pos[i].im > y_max) y_max = pos[i].im;
  }
  print (pos);
  debug_out (0, "Done.");
}

void compute_speed (void)
{
  cplx_t tmp;
  double v, ratio = min (((double) (width - 2 * BORDER)) / (x_max - x_min),
			 ((double) (height - 2 * BORDER)) / (y_max - y_min));
  int i;
  debug_in (0, "Computing speeds...");

  v_max = 0.0;
  for (i = 0; i < n; i++) {
    /* speed <- speed + accel * dt */
    tmp = accel[i];
    cr_mul (&tmp, DT);
    cc_add (speed + i, &tmp);
    if ((v = c_norm (speed + i)) > v_max)
      v_max = v;
  }
  fprintf (stderr, "max speed = %g\n", v_max * DT * ratio);
  print (speed);
  debug_out (0, "Done.");
}



void compute_accel (void)
{
  int i, j, v;
  cplx_t F;
  double dist;
  debug_in (0, "Computing accelerations...");

  for (i = 0; i < n; i++) {
    accel[i].re = accel[i].im = 0.0;
  }

  /* compute node repulses for each _other_ node*/
  debug_in (0, "Computing repulses...");
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++) {
      if (i == j) continue;
      F = pos[i];
      cc_sub (&F, pos + j);
      dist = c_norm (&F);
      if (dist == 0.0) continue;
      cr_mul (&F, node_repulse (dist) / dist);
      cc_add (accel + i, &F);
    }
  print (accel);
  debug_out (0, "Done");
  /* compute edge attracts */
  debug_in (0, "Computing attracts...");
  for (i = 0; i < n; i++)
    for (j = 0; j < d; j++) {
      if ((v = neigbhoor (i, j)) == SHARP) continue;
      F = pos[v];
      cc_sub (&F, pos + i);
      dist = c_norm (&F);
      if (dist == 0.0) continue;
      cr_mul (&F, edge_attract (dist) / dist);
      cc_add (accel + i, &F);
    }
  print (accel);
  debug_out (0, "Done");

  /* compute rubbings */
  debug_in (0, "Computing rubbings...");
  for (i = 0; i < n; i++) {
    cplx_t V = speed[i];
    F = V;
    dist = c_norm (&F);
    if (dist == 0.0) continue;
    if (dist < RUBBING) {
      accel[i].re = accel[i].im = 0.0;
      continue;
    }
    cr_mul (&V, - RUBBING / dist);
    cr_mul (&F, rubbing (dist) / dist);
    cc_sub (accel + i, &F);
    cc_add (accel + i, &V);
  }
  print (accel);
  debug_out (0, "Done");
  debug_out (0, "Done.");
}

/* unit -> unit */
value caml_step (value nothing)
{
  CAMLparam1 (nothing);
  debug_in (0, "Computing step...");

  compute_accel ();
  compute_speed ();
  compute_pos ();
  compute_int_pos ();
  print_int_pos ();

  debug_out (0, "Done.");
  CAMLreturn (Unit);
}

value positions (value graph, value w, value h)
{
  CAMLparam3 (graph, w, h);

  G = Field (graph, 0);
  n = Field (graph, 1);
  d = Field (graph, 2);

  pos = malloc (sizeof (cplx_t) * n);
  assert (pos != NULL);
  speed = malloc (sizeof (cplx_t) * n);
  assert (speed != NULL);
  accel = malloc (sizeof (cplx_t) * n);
  assert (accel != NULL);

  CAMLreturn (Unit);
}





